#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/linkage.h>
#include <linux/rbtree.h>
#include <linux/spinlock.h>
#include <linux/spinlock_types.h>
#include <linux/export.h>
#include <linux/sched.h>
#include <linux/sched/signal.h>
#include <linux/sched/rt.h>
#include <linux/sched/deadline.h>
#include <linux/sched/wake_q.h>
#include <linux/sched/debug.h>
#include <linux/timer.h>
#include <linux/rwsem.h>
#include "pirwsem_c.h"

//遍历优先级链的最大次数
int max_depth_lock = 1024;

static void
pi_rwsem_set_owner(struct pi_rwsem *lock, struct task_struct *owner)
{
    unsigned long val = (unsigned long)owner;

    if (pi_rwsem_has_waiters(lock))
        val |= PI_RWSEM_HAS_WAITERS;

    lock->owner = (struct task_struct *)val;
}

static inline void clear_pi_rwsem_waiters(struct pi_rwsem *lock)
{
    lock->owner = (struct task_struct *)((unsigned long)lock->owner & ~PI_RWSEM_HAS_WAITERS);
}

static void fixup_pi_rwsem_waiters(struct pi_rwsem *lock)
{
    unsigned long owner, *p = (unsigned long *)&lock->owner;

    if (pi_rwsem_has_waiters(lock))
        return;
    owner = READ_ONCE(*p);
    if (owner & PI_RWSEM_HAS_WAITERS)
        WRITE_ONCE(*p, owner & ~PI_RWSEM_HAS_WAITERS);
}

/*
 * We can speed up the acquire/release, if there's no debugging state to be
 * set up.
 */
// #ifndef CONFIG_DEBUG_PI_RWSEMES
#define pi_rwsem_cmpxchg_relaxed(l, c, n) (cmpxchg_relaxed(&l->owner, c, n) == c)
#define pi_rwsem_cmpxchg_acquire(l, c, n) (cmpxchg_acquire(&l->owner, c, n) == c)
#define pi_rwsem_cmpxchg_release(l, c, n) (cmpxchg_release(&l->owner, c, n) == c)

// 标记等待mutex的任务
static inline void mark_pi_rwsem_waiters(struct pi_rwsem *lock)
{
    unsigned long owner, *p = (unsigned long *)&lock->owner;

    do
    {
        owner = *p;
    } while (cmpxchg_relaxed(p, owner,
                             owner | PI_RWSEM_HAS_WAITERS) != owner);
}

// 安全解锁mutex
static inline bool unlock_pi_rwsem_safe(struct pi_rwsem *lock,
					unsigned long flags)
	__releases(lock->wait_lock)
{
	struct task_struct *owner = pi_rwsem_owner(lock);

	clear_pi_rwsem_waiters(lock);
	raw_spin_unlock_irqrestore(&lock->wait_lock, flags);
	return pi_rwsem_cmpxchg_release(lock, owner, NULL);
}

// tesk 2 task_to_waiter
#define task_to_waiter(p)	\
	&(struct rt_mutex_waiter){ .prio = (p)->prio, .deadline = (p)->dl.deadline }

static inline int
pi_rwsem_waiter_less(struct rt_mutex_waiter *left,
		     struct rt_mutex_waiter *right)
{
	if (left->prio < right->prio)
		return 1;

	/*
	 * If both waiters have dl_prio(), we check the deadlines of the
	 * associated tasks.
	 * If left waiter has a dl_prio(), and we didn't return 1 above,
	 * then right waiter has a dl_prio() too.
	 */
	if (dl_prio(left->prio))
		return dl_time_before(left->deadline, right->deadline);

	return 0;
}

// 比较两个waiter任务优先级是否相等
static inline int
pi_rwsem_waiter_equal(struct rt_mutex_waiter *left,
		      struct rt_mutex_waiter *right)
{
	if (left->prio != right->prio)
		return 0;

	if (dl_prio(left->prio))
		return left->deadline == right->deadline;

	return 1;
}

// 将waiter持有的mutex插入RBtree
static void
pi_rwsem_enqueue(struct pi_rwsem *lock, struct rt_mutex_waiter *waiter)
{
	struct rb_node **link = &lock->waiters.rb_root.rb_node;
	struct rb_node *parent = NULL;
	struct rt_mutex_waiter *entry;
	bool leftmost = true;

	while (*link) {
		parent = *link;
		entry = rb_entry(parent, struct rt_mutex_waiter, tree_entry);
		if (pi_rwsem_waiter_less(waiter, entry)) {
			link = &parent->rb_left;
		} else {
			link = &parent->rb_right;
			leftmost = false;
		}
	}

	rb_link_node(&waiter->tree_entry, parent, link);
	rb_insert_color_cached(&waiter->tree_entry, &lock->waiters, leftmost);
}
// 从RBtree移除waiter mutex
static void
pi_rwsem_dequeue(struct pi_rwsem *lock, struct rt_mutex_waiter *waiter)
{
	if (RB_EMPTY_NODE(&waiter->tree_entry))
		return;

	rb_erase_cached(&waiter->tree_entry, &lock->waiters);
	RB_CLEAR_NODE(&waiter->tree_entry);
}

// 将请求mutex受阻的waiter插入RBtree
static void
pi_rwsem_enqueue_pi(struct task_struct *task, struct rt_mutex_waiter *waiter)
{
	struct rb_node **link = &task->pi_waiters.rb_root.rb_node;
	struct rb_node *parent = NULL;
	struct rt_mutex_waiter *entry;
	bool leftmost = true;

	while (*link) {
		parent = *link;
		entry = rb_entry(parent, struct rt_mutex_waiter, pi_tree_entry);
		if (pi_rwsem_waiter_less(waiter, entry)) {
			link = &parent->rb_left;
		} else {
			link = &parent->rb_right;
			leftmost = false;
		}
	}

	rb_link_node(&waiter->pi_tree_entry, parent, link);
	rb_insert_color_cached(&waiter->pi_tree_entry, &task->pi_waiters, leftmost);
}

// 将请求mutex受阻的waiter移出RBtree
static void
pi_rwsem_dequeue_pi(struct task_struct *task, struct rt_mutex_waiter *waiter)
{
	if (RB_EMPTY_NODE(&waiter->pi_tree_entry))
		return;

	rb_erase_cached(&waiter->pi_tree_entry, &task->pi_waiters);
	RB_CLEAR_NODE(&waiter->pi_tree_entry);
}

// 调整任务优先级
static void pi_rwsem_adjust_prio(struct task_struct *p)
{
	struct task_struct *pi_task = NULL;

	lockdep_assert_held(&p->pi_lock);

	if (task_has_pi_waiters(p))
		pi_task = task_top_pi_waiter(p)->task;

	rt_mutex_setprio(p, pi_task);
}

// 检测是否发生死锁
static bool pi_rwsem_cond_detect_deadlock(struct rt_mutex_waiter *waiter,
					  enum pirwsem_chainwalk chwalk)
{
	return debug_pi_rwsem_detect_deadlock(waiter, chwalk);
}

// 获取阻塞当前任务的mutex
static inline struct pi_rwsem *task_blocked_on_lock(struct task_struct *p)
{
	return p->pi_blocked_on ? p->pi_blocked_on->lock : NULL;
}

// 调整优先级链
static int pi_rwsem_adjust_prio_chain(struct task_struct *task,
				      enum pirwsem_chainwalk chwalk,
				      struct pi_rwsem *orig_lock,
				      struct pi_rwsem *next_lock,
				      struct rt_mutex_waiter *orig_waiter,
				      struct task_struct *top_task)
{
	struct rt_mutex_waiter *waiter, *top_waiter = orig_waiter;
	struct rt_mutex_waiter *prerequeue_top_waiter;
	int ret = 0, depth = 0;
	struct pi_rwsem *lock;
	bool detect_deadlock;
	bool requeue = true;
    // 检测死锁
	detect_deadlock = pi_rwsem_cond_detect_deadlock(orig_waiter, chwalk);

 again:
	if (++depth > max_depth_lock) {
		static int prev_max;

		if (prev_max != max_depth_lock) {
			prev_max = max_depth_lock;
			printk(KERN_WARNING "Maximum lock depth %d reached "
			       "task: %s (%d)\n", max_depth_lock,
			       top_task->comm, task_pid_nr(top_task));
		}
		put_task_struct(task);

		return -EDEADLK;
	}

 retry:
	raw_spin_lock_irq(&task->pi_lock);
	waiter = task->pi_blocked_on;

	// 检查是否到达链尾，或者链的状态在释放锁之后发生了变化。
	if (!waiter)
		goto out_unlock_pi;

	// 检查 orig_waiter 的状态
	if (orig_waiter && !pi_rwsem_owner(orig_lock))
		goto out_unlock_pi;

	// 检测mutex链是否发生变化
	if (next_lock != waiter->lock)
		goto out_unlock_pi;

	// 如果任务没有等待者，则退出
	if (top_waiter) {
		if (!task_has_pi_waiters(task))
			goto out_unlock_pi;

		if (top_waiter != task_top_pi_waiter(task)) {
			if (!detect_deadlock)
				goto out_unlock_pi;
			else
				requeue = false;
		}
	}

	// 如果等待者的优先级与任务的优先级相同，则无需进一步调整优先级
	if (pi_rwsem_waiter_equal(waiter, task_to_waiter(task))) {
		if (!detect_deadlock)
			goto out_unlock_pi;
		else
			requeue = false;
	}

	// 获取下一个锁
	lock = waiter->lock;
	// 释放任务的 pi_lock 锁和等待锁
	if (!raw_spin_trylock(&lock->wait_lock)) {
		raw_spin_unlock_irq(&task->pi_lock);
		cpu_relax();
		goto retry;
	}

	if (lock == orig_lock || pi_rwsem_owner(lock) == top_task) {
		raw_spin_unlock(&lock->wait_lock);
		ret = -EDEADLK;
		goto out_unlock_pi;
	}

	if (!requeue) {
		raw_spin_unlock(&task->pi_lock);
		put_task_struct(task);

		if (!pi_rwsem_owner(lock)) {
			raw_spin_unlock_irq(&lock->wait_lock);
			return 0;
		}

		task = pi_rwsem_owner(lock);
		get_task_struct(task);
		raw_spin_lock(&task->pi_lock);

		next_lock = task_blocked_on_lock(task);
		top_waiter = pi_rwsem_top_waiter(lock);

		raw_spin_unlock(&task->pi_lock);
		raw_spin_unlock_irq(&lock->wait_lock);

		if (!next_lock)
			goto out_put_task;
		goto again;
	}

	prerequeue_top_waiter = pi_rwsem_top_waiter(lock);
	pi_rwsem_dequeue(lock, waiter);

	waiter->prio = task->prio;
	waiter->deadline = task->dl.deadline;

	pi_rwsem_enqueue(lock, waiter);

	raw_spin_unlock(&task->pi_lock);
	put_task_struct(task);

	if (!pi_rwsem_owner(lock)) {
		if (prerequeue_top_waiter != pi_rwsem_top_waiter(lock))
			wake_up_process(pi_rwsem_top_waiter(lock)->task);
		raw_spin_unlock_irq(&lock->wait_lock);
		return 0;
	}

	task = pi_rwsem_owner(lock);
	get_task_struct(task);
	raw_spin_lock(&task->pi_lock);

	if (waiter == pi_rwsem_top_waiter(lock)) {
		pi_rwsem_dequeue_pi(task, prerequeue_top_waiter);
		pi_rwsem_enqueue_pi(task, waiter);
		pi_rwsem_adjust_prio(task);

	} else if (prerequeue_top_waiter == waiter) {
		pi_rwsem_dequeue_pi(task, waiter);
		waiter = pi_rwsem_top_waiter(lock);
		pi_rwsem_enqueue_pi(task, waiter);
		pi_rwsem_adjust_prio(task);
	} else {

	}

	next_lock = task_blocked_on_lock(task);
	top_waiter = pi_rwsem_top_waiter(lock);

	raw_spin_unlock(&task->pi_lock);
	raw_spin_unlock_irq(&lock->wait_lock);

	if (!next_lock)
		goto out_put_task;

	if (!detect_deadlock && waiter != top_waiter)
		goto out_put_task;

	goto again;

 out_unlock_pi:
	raw_spin_unlock_irq(&task->pi_lock);
 out_put_task:
	put_task_struct(task);

	return ret;
}

// 尝试获取pi_rwsem
static int try_to_take_pi_rwsem(struct pi_rwsem *lock, struct task_struct *task,
				struct rt_mutex_waiter *waiter)
{
	lockdep_assert_held(&lock->wait_lock);
	mark_pi_rwsem_waiters(lock);

	if (pi_rwsem_owner(lock))
		return 0;

	if (waiter) {
		if (waiter != pi_rwsem_top_waiter(lock))
			return 0;
		pi_rwsem_dequeue(lock, waiter);

	} else {
		if (pi_rwsem_has_waiters(lock)) {
			if (!pi_rwsem_waiter_less(task_to_waiter(task),
						  pi_rwsem_top_waiter(lock)))
				return 0;
		} else {
			goto takeit;
		}
	}

	raw_spin_lock(&task->pi_lock);
	task->pi_blocked_on = NULL;

	if (pi_rwsem_has_waiters(lock))
		pi_rwsem_enqueue_pi(task, pi_rwsem_top_waiter(lock));
	raw_spin_unlock(&task->pi_lock);

takeit:
	pi_rwsem_set_owner(lock, task);

	return 1;
}

// 任务获取pi_rwsem受阻
static int task_blocks_on_pi_rwsem(struct pi_rwsem *lock,
				   struct rt_mutex_waiter *waiter,
				   struct task_struct *task,
				   enum pirwsem_chainwalk chwalk)
{
	struct task_struct *owner = pi_rwsem_owner(lock);
	struct rt_mutex_waiter *top_waiter = waiter;
	struct pi_rwsem *next_lock;
	int chain_walk = 0, res;

	lockdep_assert_held(&lock->wait_lock);

	if (owner == task)
		return -EDEADLK;

	raw_spin_lock(&task->pi_lock);
	waiter->task = task;
	waiter->lock = lock;
	waiter->prio = task->prio;
	waiter->deadline = task->dl.deadline;

	if (pi_rwsem_has_waiters(lock))
		top_waiter = pi_rwsem_top_waiter(lock);
	pi_rwsem_enqueue(lock, waiter);

	task->pi_blocked_on = waiter;

	raw_spin_unlock(&task->pi_lock);

	if (!owner)
		return 0;

	raw_spin_lock(&owner->pi_lock);
	if (waiter == pi_rwsem_top_waiter(lock)) {
		pi_rwsem_dequeue_pi(owner, top_waiter);
		pi_rwsem_enqueue_pi(owner, waiter);

		pi_rwsem_adjust_prio(owner);
		if (owner->pi_blocked_on)
			chain_walk = 1;
	} else if (pi_rwsem_cond_detect_deadlock(waiter, chwalk)) {
		chain_walk = 1;
	}

	next_lock = task_blocked_on_lock(owner);
	raw_spin_unlock(&owner->pi_lock);

	if (!chain_walk || !next_lock)
		return 0;

	get_task_struct(owner);
	raw_spin_unlock_irq(&lock->wait_lock);

	res = pi_rwsem_adjust_prio_chain(owner, chwalk, lock,
					 next_lock, waiter, task);

	raw_spin_lock_irq(&lock->wait_lock);

	return res;
}

// 标记即将唤醒的waiter
static void mark_wakeup_next_waiter(struct wake_q_head *wake_q,
				    struct pi_rwsem *lock)
{
	struct rt_mutex_waiter *waiter;

	raw_spin_lock(&current->pi_lock);

	waiter = pi_rwsem_top_waiter(lock);

	pi_rwsem_dequeue_pi(current, waiter);
	pi_rwsem_adjust_prio(current);

	lock->owner = (void *) PI_RWSEM_HAS_WAITERS;

	preempt_disable();
	wake_q_add(wake_q, waiter->task);
	raw_spin_unlock(&current->pi_lock);
}

// 移除waiter
static void remove_waiter(struct pi_rwsem *lock,
			  struct rt_mutex_waiter *waiter)
{
	bool is_top_waiter = (waiter == pi_rwsem_top_waiter(lock));
	struct task_struct *owner = pi_rwsem_owner(lock);
	struct pi_rwsem *next_lock;

	lockdep_assert_held(&lock->wait_lock);

	raw_spin_lock(&current->pi_lock);
	pi_rwsem_dequeue(lock, waiter);
	current->pi_blocked_on = NULL;
	raw_spin_unlock(&current->pi_lock);

	if (!owner || !is_top_waiter)
		return;

	raw_spin_lock(&owner->pi_lock);

	pi_rwsem_dequeue_pi(owner, waiter);

	if (pi_rwsem_has_waiters(lock))
		pi_rwsem_enqueue_pi(owner, pi_rwsem_top_waiter(lock));

	pi_rwsem_adjust_prio(owner);
	next_lock = task_blocked_on_lock(owner);
	raw_spin_unlock(&owner->pi_lock);

	if (!next_lock)
		return;

	get_task_struct(owner);

	raw_spin_unlock_irq(&lock->wait_lock);

	pi_rwsem_adjust_prio_chain(owner, PI_RWSEM_MIN_CHAINWALK, lock,
				   next_lock, NULL, current);

	raw_spin_lock_irq(&lock->wait_lock);
}

//检查优先级继承链
void pi_rwsem_adjust_pi(struct task_struct *task)
{
	struct rt_mutex_waiter *waiter;
	struct pi_rwsem *next_lock;
	unsigned long flags;

	raw_spin_lock_irqsave(&task->pi_lock, flags);

	waiter = task->pi_blocked_on;
	if (!waiter || pi_rwsem_waiter_equal(waiter, task_to_waiter(task))) {
		raw_spin_unlock_irqrestore(&task->pi_lock, flags);
		return;
	}
	next_lock = waiter->lock;
	raw_spin_unlock_irqrestore(&task->pi_lock, flags);
	get_task_struct(task);

	pi_rwsem_adjust_prio_chain(task, PI_RWSEM_MIN_CHAINWALK, NULL,
				   next_lock, NULL, task);
}

void pi_rwsem_init_waiter(struct rt_mutex_waiter *waiter)
{
	RB_CLEAR_NODE(&waiter->pi_tree_entry);
	RB_CLEAR_NODE(&waiter->tree_entry);
	waiter->task = NULL;
}

// 执行等待-唤醒-尝试获取循环
static int __sched
__pi_rwsem_slowlock(struct pi_rwsem *lock, int state,
		    struct hrtimer_sleeper *timeout,
		    struct rt_mutex_waiter *waiter)
{
	int ret = 0;

	for (;;) {
		if (try_to_take_pi_rwsem(lock, current, waiter))
			break;
		if (likely(state == TASK_INTERRUPTIBLE)) {
			if (signal_pending(current))
				ret = -EINTR;
			if (timeout && !timeout->task)
				ret = -ETIMEDOUT;
			if (ret)
				break;
		}

		raw_spin_unlock_irq(&lock->wait_lock);

		schedule();

		raw_spin_lock_irq(&lock->wait_lock);
		set_current_state(state);
	}
	__set_current_state(TASK_RUNNING);
	return ret;
}

static void pi_rwsem_handle_deadlock(int res, int detect_deadlock,
				     struct rt_mutex_waiter *w)
{
	if (res != -EDEADLOCK || detect_deadlock)
		return;

	pi_rwsem_print_deadlock(w);
	while (1) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule();
	}
}

//持续等待锁
static int __sched
pi_rwsem_slowlock(struct pi_rwsem *lock, int state,
		  struct hrtimer_sleeper *timeout,
		  enum pirwsem_chainwalk chwalk)
{
	struct rt_mutex_waiter waiter;
	unsigned long flags;
	int ret = 0;

	pi_rwsem_init_waiter(&waiter);
	raw_spin_lock_irqsave(&lock->wait_lock, flags);

	if (try_to_take_pi_rwsem(lock, current, NULL)) {
		raw_spin_unlock_irqrestore(&lock->wait_lock, flags);
		return 0;
	}

	set_current_state(state);

	if (unlikely(timeout))
		hrtimer_start_expires(&timeout->timer, HRTIMER_MODE_ABS);

	ret = task_blocks_on_pi_rwsem(lock, &waiter, current, chwalk);

	if (likely(!ret))
		ret = __pi_rwsem_slowlock(lock, state, timeout, &waiter);

	if (unlikely(ret)) {
		__set_current_state(TASK_RUNNING);
		remove_waiter(lock, &waiter);
		pi_rwsem_handle_deadlock(ret, chwalk, &waiter);
	}

	fixup_pi_rwsem_waiters(lock);

	raw_spin_unlock_irqrestore(&lock->wait_lock, flags);

	if (unlikely(timeout))
		hrtimer_cancel(&timeout->timer);

	return ret;
}


//释放mutex
static bool __sched pi_rwsem_slowunlock(struct pi_rwsem *lock,
					struct wake_q_head *wake_q)
{
	unsigned long flags;
	raw_spin_lock_irqsave(&lock->wait_lock, flags);

	while (!pi_rwsem_has_waiters(lock)) {
		if (unlock_pi_rwsem_safe(lock, flags) == true)
			return false;
		raw_spin_lock_irqsave(&lock->wait_lock, flags);
	}

	mark_wakeup_next_waiter(wake_q, lock);
	raw_spin_unlock_irqrestore(&lock->wait_lock, flags);

	return true;
}

// 快速获取pi_rwsem，否则持续等待pi_rwsem
static inline int
pi_rwsem_fastlock(struct pi_rwsem *lock, int state,
		  int (*slowfn)(struct pi_rwsem *lock, int state,
				struct hrtimer_sleeper *timeout,
				enum pirwsem_chainwalk chwalk))
{
	if (likely(pi_rwsem_cmpxchg_acquire(lock, NULL, current)))
		return 0;

	return slowfn(lock, state, NULL, PI_RWSEM_MIN_CHAINWALK);
}

// 唤醒顶部waiter并重新启用抢占。
void pi_rwsem_postunlock(struct wake_q_head *wake_q)
{
	wake_up_q(wake_q);
	preempt_enable();
}

static inline void
pi_rwsem_fastunlock(struct pi_rwsem *lock,
		    bool (*slowfn)(struct pi_rwsem *lock,
				   struct wake_q_head *wqh))
{
	DEFINE_WAKE_Q(wake_q);

	if (likely(pi_rwsem_cmpxchg_release(lock, current, NULL)))
		return;

	if (slowfn(lock, &wake_q))
		pi_rwsem_postunlock(&wake_q);
}

static inline void __pi_down_write(struct pi_rwsem *lock, unsigned int subclass)
{
	might_sleep();

	mutex_acquire(&lock->dep_map, subclass, 0, _RET_IP_);
	pi_rwsem_fastlock(lock, TASK_UNINTERRUPTIBLE, pi_rwsem_slowlock);
}

// 获取pi写信号量
void __sched pi_down_write(struct pi_rwsem *lock)
{
	__pi_down_write(lock, 0);
}
EXPORT_SYMBOL_GPL(pi_down_write);

// 释放pi写信号量
void __sched pi_up_write(struct pi_rwsem *lock)
{
	mutex_release(&lock->dep_map, 1, _RET_IP_);
	pi_rwsem_fastunlock(lock, pi_rwsem_slowunlock);
}
EXPORT_SYMBOL_GPL(pi_up_write);

// 获取pi读信号量
void __sched pi_down_read(struct rw_semaphore *lock)
{
	down_read(lock);
}
EXPORT_SYMBOL(pi_down_read);

// 释放pi读信号量
void __sched pi_up_read(struct rw_semaphore *lock)
{
	up_read(lock);
}
EXPORT_SYMBOL(pi_up_read);



// 初始化pi_rwsem
void __pi_rwsem_init(struct rw_semaphore *lock, const char *name,
		  struct lock_class_key *key)
{
	lock->count = 0;
	raw_spin_lock_init(&lock->wait_lock);
	INIT_LIST_HEAD(&lock->wait_list);
	lock->waiters = RB_ROOT_CACHED;
}
EXPORT_SYMBOL(__pi_rwsem_init);

/*
** 初始化模块
*/
static int __init pi_rwsem_init(void)
{
    printk("Device Driver Insert...Done!!!\n");
    return 0;
}

/*
** 卸载模块
*/
static void __exit pi_rwsem_exit(void)
{
    printk("Device Driver Remove...Done!!\n");
}

module_init(pi_rwsem_init);
module_exit(pi_rwsem_exit);
MODULE_LICENSE("GPL");